--[[
    Coconuts by LordFarlander
    Version 2.5
    $Revision: 80185 $

    Used to be called Mount.
]]--

--[[
Copyright (c) 2008, LordFarlander
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
]]--

local L = LibStub( "AceLocale-3.0" ):GetLocale( "Coconuts" );

_G["BINDING_NAME_CLICK " .. "Coconuts" .. "_Button:LeftButton"] = L["Summon Mount"];
BINDING_HEADER_COCONUTSHEADER = L["Coconuts"];

if( not LibStub( "LibSpecialEvents-Mount-4.0", true ) ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibSpecialEvents-Mount-4.0" ) );
    return;
end--if

if( not LibStub( "LibSpecialEvents-LearnSpell-4.0", true ) ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibSpecialEvents-LearnSpell-4.0" ) );
    return;
end--if

if( not LibStub( "LibSpecialEvents-Skill-4.0", true ) ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibSpecialEvents-Skill-4.0" ) );
    return;
end--if

local LibLordFarlander = LibStub( "LibLordFarlander-2.0", true );
if( not LibLordFarlander ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-2.0" ) );
    return;
end--if

local ItemList = LibStub( "LibLordFarlander-ItemList-2.0", true );
if( not ItemList ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-ItemList-2.0" ) );
    return;
end--if

if( not LibStub( "LibLordFarlander-UI-2.0", true ) ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-UI-2.0" ) );
    return;
end--if

local PetAndMountDatabase = LibStub( "LibPetAndMountDatabase-1.1", true );
if( not PetAndMountDatabase ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibPetAndMountDatabase-1.1" ) );
    return;
end--if

Coconuts = LibStub( "AceAddon-3.0" ):NewAddon( "Coconuts", "AceEvent-3.0", "AceTimer-3.0", "AceConsole-3.0" );
Coconuts.version = { major = 2, minor = 5, revision = tonumber( ("$Rev: 80185 $"):match( "(%d+)" ) ) };

Coconuts.DEBUG  = false;            --true = Debugging Mode
Coconuts.dprint = LibLordFarlander.dprint;
Coconuts.button = nil;
Coconuts.ui = { currentZone = "", currentFavorite = 0, Favorites = {}, currentType = 0, Types = {}, };

Coconuts.dirtyBags = {};

--DO NOT CHANGE THESE VALUES
Coconuts.Ground = { Best = "", NextBest = "", Passenger = { Best = "", NextBest = "", Sets = {}, }, Sets = {}, };
Coconuts.Flying = { Best = "", NextBest = "", Passenger = { Best = "", NextBest = "", Sets = {}, }, Sets = {}, };

Coconuts.CurrentGround     = nil; --Ground mount on the Button
Coconuts.CurrentFlying     = nil; --Flying mount on the Button
Coconuts.CurrentMount      = nil; --Current mount
Coconuts.CurrentMountItem  = nil; --Current mount modifier item
Coconuts.NextCurrentGround = nil; --Ground mount on the Button
Coconuts.NextCurrentFlying = nil; --Flying mount on the Button
Coconuts.CurrentGroundForm = nil;
Coconuts.CurrentFlyingForm = nil;
Coconuts.CurrentAquaticForm = nil;
Coconuts.CurrentGroundPassenger     = nil;
Coconuts.CurrentFlyingPassenger     = nil;
Coconuts.NextCurrentGroundPassenger = nil;
Coconuts.NextCurrentFlyingPassenger = nil;
Coconuts.TempleOfAhnQiraj  = nil; --Coconuts allowed in Temple Of Ahn'Qiraj
Coconuts.ReadyToInitialize = false;
Coconuts.Mounted           = false;
Coconuts.Swimming          = false;
Coconuts.Dirty             = false;
Coconuts.Mode              = "";
Coconuts.RidingSkill       = nil;
Coconuts.Loaded            = false;

function Coconuts:OnInitialize()
    local AceConfigReg = LibStub( "AceConfigRegistry-3.0" );
    local AceConfigDialog = LibStub( "AceConfigDialog-3.0" );
    local defaults = {
            profile = {
                FavorFlying = true,
                FavorPassenger = true,
                ShowModifierItems = true,
                SuggestOnlySoulbound = true,
                ModiferDismountFlying = true,
                SuggestForms = true,
                DefaultToFormInCombat = true,
                FlyingModifier = "Alt",
                NextBestModifier = "Shift",
                PassengerModifier = "Control",
                MountItemModifier = "Shift",
                -- List of favorite mounts, these will be picked more often
                -- stored as [itemID]=percentage
                Favorites = {},
                -- List of mounts per zone, if the mounts are in the player's bags or spells they will be picked
                -- when in the indicated zone
                -- stored as ["Localized zone name"]={ "type"=itemID, ... }
                ZoneFavorites = {},
                Button = LibStub( "LibLordFarlander-UI-2.0" ).GetButtonSettingsDefaults(),
            },
        };
    local buttonOptions = LibStub( "LibLordFarlander-UI-2.0" ).GetButtonOptionGroup( self, {
            name = L["Button"],
            desc = L["Options relating to the button."],
            yes = L["yes"],
            no = L["no"],
            show = L["showing"],
            hide = L["hiding"],
            locked = L["locked"],
            unlocked = L["unlocked"],
            toggleButton = L["toggleButton"],
            toggleButton_name = L["Show button"],
            toggleButton_desc = L["Show/hide the button."],
            autoHideButton = L["autoHideButton"],
            autoHideButton_name = L["Automatically hide button"],
            autoHideButton_desc = L["Hide the button when you can't mount?"],
            lockButton = L["lockButton"],
            lockButton_name = L["Lock button position"],
            lockButton_desc = L["Lock/unlock the position of the button."],
            alpha = L["alpha"],
            alpha_name = L["Alpha of button"],
            alpha_desc = L["Sets the alpha value that the button has normally."],
            fadeOut = L["fadeOut"],
            fadeOut_name = L["Fade out"],
            fadeOut_desc = L["Fade the button out when the cursor is not over it?"],
            fadeOutInsteadHide = L["fadeOutInsteadHide"],
            fadeOutInsteadHide_name = L["Fade out instead of hiding"],
            fadeOutInsteadHide_desc = L["Fade the button out instead of hiding it?"],
            fadeOutCancelInCombat = L["fadeOutCancelInCombat"],
            fadeOutCancelInCombat_name = L["Cancel fade out if in combat"],
            fadeOutCancelInCombat_desc = L["Should the button not fade out if you're in combat?"],
            fadeOutCancelOnShift = L["fadeOutCancelOnShift"],
            fadeOutCancelOnShift_name = L["Cancel fade out if holding Shift"],
            fadeOutCancelOnShift_desc = L["Should the button not fade out if you're holding Shift?"],
            fadeOutCancelOnCtrl = L["fadeOutCancelOnCtrl"],
            fadeOutCancelOnCtrl_name = L["Cancel fade out if holding Ctrl"],
            fadeOutCancelOnCtrl_desc = L["Should the button not fade out if you're holding Ctrl?"],
            fadeOutCancelOnAlt = L["fadeOutCancelOnAlt"],
            fadeOutCancelOnAlt_name = L["Cancel fade out if holding Alt"],
            fadeOutCancelOnAlt_desc = L["Should the button not fade out if you're holding Alt?"],
            fadeOutTime = L["fadeOutTime"],
            fadeOutTime_name = L["Fade out time"],
            fadeOutTime_desc = L["Sets how long the button takes to fade out."],
            fadeOutDelay = L["fadeOutDelay"],
            fadeOutDelay_name = L["Fade out delay"],
            fadeOutDelay_desc = L["Sets the delay before the button starts to fade."],
            fadeOutAlpha = L["fadeOutAlpha"];
            fadeOutAlpha_name = L["Alpha after fadeout"],
            fadeOutAlpha_desc = L["Sets the alpha value that the button has after faded out."],
        }, 1 );
    local favoritesOptions = {
        type = "group",
        name = L["Favorites"],
        desc = L["Mounts that should be picked more often if you have them."],
        handler = self,
        args = {
            favorites = {
                type = "select",
                order = 1,
                name = L["Favorites"],
                desc = L["Mounts that should be picked more often if you have them."],
                width = "full",
                cmdHidden = true,
                values = function( info )
                             info.handler.ui.Favorites = {};

                             local Favorites = info.handler.ui.Favorites;
                             local firstItem = 0;
                             local used = {};

                             for _, mount in pairs( info.handler:GetAllMountsList() ) do
                                 local id = mount:GetID();

                                 used[id] = true;
                                 table.insert( Favorites, id );
                                 if( type( info.handler.db.profile.Favorites[id] ) == "boolean" ) then
                                     info.handler.db.profile.Favorites[id] = 2.0;
                                 end--if
                                 if( type( info.handler.db.profile.Favorites[id] ) == "nil" ) then
                                     info.handler.db.profile.Favorites[id] = 1.0;
                                 end--if
                             end--for
                             for id, howOften in pairs( info.handler.db.profile.Favorites ) do
                                 if( not used[id] ) then
                                     table.insert( Favorites, id );
                                     used[id] = true;
                                 end--if
                             end--if
                             table.sort( Favorites, function( a, b ) return ( ( a < 0 ) and GetSpellInfo( a * -1 ) or GetItemInfo( a ) ) < ( ( b < 0 ) and GetSpellInfo( b * -1 ) or GetItemInfo( b ) ) end );
                             Favorites = {};
                             for index, mount in pairs( info.handler.ui.Favorites ) do
                                 local sName, sTexture;
                                 local sHowOften = L["default"];
                                 local howOften = info.handler.db.profile.Favorites[mount];

                                 if( mount < 0 ) then
                                     sName, _, sTexture, _, _, _, _, _, _ = GetSpellInfo( mount * -1 );
                                 else
                                     sName, _, _, _, _, _, _, _, _, sTexture = GetItemInfo( mount );
                                 end--if
                                 if( howOften == 0.0 ) then
                                     sHowOften = L["never"];
                                 elseif( howOften ~= 1.0 ) then
                                     sHowOften = L["%.1f times as often"]:format( howOften );
                                 end--if                                 
                                 if( sTexture ) then
                                     Favorites[index] = ("|T%s:0|t %s |cFFFF0000[%s]|r"):format( sTexture, sName, sHowOften );
                                 end--if
                                 if( firstItem == 0 ) then
                                     firstItem = index;
                                 end--if
                             end--for
                             if( ( not Favorites[info.handler.ui.currentFavorite] ) or ( info.handler.ui.currentFavorite == 0 ) ) then
                                 info.handler.ui.currentFavorite = firstItem;
                             end--if
                             return Favorites;
                         end,
                get = function( info )
                          return info.handler.ui.currentFavorite;
                      end,
                set = function( info, val )
                          info.handler.ui.currentFavorite = val;
                      end,
            },
            frequency = {
                type = "range",
                order = 2,
                name = L["Frequency"],
                cmdHidden = true,
                min = 0.0,
                max = 5.0,
                bigStep = 0.1,
                isPercent = true,
                desc = L["Set how often the mount gets picked.  2 is twice as often and 0 is never."];
                get = function( info )
                          return info.handler.db.profile.Favorites[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] or 1.0;
                      end,
                set = function( info, val )
                          info.handler.db.profile.Favorites[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = val;
                      end,
            },

            [L["list"]] = {
                type = "execute",
                order = 3,
                guiHidden = true,
                name = L["List all favorite mounts"],
                desc = L["Show a list with all of your favorite mounts."],
                func = function( info )
                           info.handler:Print( L["List of your favorite mounts:"] );
                           for mount, howOften in pairs( info.handler.db.profile.Favorites ) do
                               if( howOften ~= 1.0 ) then
                                   local link = ( mount > 0 ) and LibLordFarlander.GetItemLink( mount ) or GetSpellLink( mount * -1 );

                                   if( mount > 0 ) then
                                       _, _, _, _, _, _, _, _, _, sTexture = GetItemInfo( mount );
                                   else
                                       _, _, sTexture, _, _, _, _, _, _ = GetSpellInfo( mount * -1 );
                                   end--if
                                   if( howOften == 0.0 ) then
                                       howOften = L["never"];
                                   else
                                       howOften = L["%.1f times as often"]:format( howOften );
                                   end--if

                                   if( sTexture ) then
                                       info.handler:Print( ("|T%s:0|t %s |cFFFF0000[%s]|r"):format( sTexture, link or tostring( mount ), howOften ) );
                                   else
                                       info.handler:Print( ("%s"):format( link or tostring( mount ) ) );
                                   end--if
                               end--if
                           end--for
                       end,
            },
        },
    };
    local zoneOptions =  {
        type = "group",
        name = L["Zone Favorites"],
        desc = L["Set a mount to use when in a certain zone."],
        handler = self,
        args = {
            zones = {
                type = "select",
                order = 1,
                name = L["Zones"],
                desc = L["Set a mount to use when in a certain zone."],
                width = "full",
                cmdHidden = true,
                values = function( info )
                             local zones = {};
                             local continents = { GetMapContinents() };

                             for continent, _ in pairs( continents ) do
                                 local continentZones = { GetMapZones( continent ) };

                                 for _, zone in pairs( continentZones ) do
                                     zones[zone] = zone;
                                 end--for
                             end--for
                             -- Check for zones we may not know about (battlegrounds for example)
                             for zone, pet in pairs( info.handler.db.profile.ZoneFavorites ) do
                                 if( not zones[zone] ) then
                                     zones[zone] = zone;
                                 end--if
                             end--for
                             if( not zones[GetRealZoneText()] ) then
                                 zones[GetRealZoneText()] = GetRealZoneText();
                             end--for
                             if( #zones ) then
                                 if( info.handler.ui.currentZone == "" ) then
                                     info.handler.ui.currentZone = GetRealZoneText();
                                 end--if
                             else
                                 info.handler.ui.currentZone = "";
                             end--if                                     
                             return zones;
                         end,
                get = function( info )
                          return info.handler.ui.currentZone;
                      end,
                set = function( info, val )
                          info.handler.ui.currentZone = val;
                          info.handler.ui.currentType = 0;
                      end,
            },
            types = {
                type = "select",
                order = 2,
                name = L["Types"],
                desc = L["Lets you pick which type of mount to set a favorite for."],
                width = "full",
                cmdHidden = true,
                values = function( info )
                             info.handler.ui.Types = {};

                             local GroundMounts = PetAndMountDatabase:GetSubsets( "Mounts.Ground" );
                             local GroundPassengerMounts = PetAndMountDatabase:GetSubsets( "Mounts.Passenger.Ground" );
                             local types = {};
                             local first = 11000;
                             local uiTypes = info.handler.ui.Types;
                             local GetSetSpeed = info.handler.GetSetSpeed;

                             if( LibLordFarlander.IsFlyableZone( info.handler.ui.currentZone, true ) ) then
                                 local FlyingMounts = PetAndMountDatabase:GetSubsets( "Mounts.Flying" );
                                 local FlyingPassengerMounts = PetAndMountDatabase:GetSubsets( "Mounts.Passenger.Flying" );

                                 for _, name in pairs( FlyingMounts ) do
                                     local speed = GetSetSpeed( name );

                                     if( speed ) then
                                         local index = speed + 10000;

                                         uiTypes[index] = name;
                                         types[index] = L["%1$i%% speed %2$s"]:format( speed, L["flying"] );
                                         if( first > index ) then
                                             first = index;
                                         end--if
                                     end--if
                                 end--for
                                 for _, name in pairs( FlyingPassengerMounts ) do
                                     local speed = GetSetSpeed( name );

                                     if( speed ) then
                                         local index = speed + 10000 + 1000;

                                         uiTypes[index] = name;
                                         types[index] = L["%1$i%% speed %2$s"]:format( speed, L["flying passenger"] );
                                         if( first > index ) then
                                             first = index;
                                         end--if
                                     end--if
                                 end--for
                             end--if
                             for _, name in pairs( GroundMounts ) do
                                 local speed = GetSetSpeed( name );

                                 if( speed ) then
                                     uiTypes[speed] = name;
                                     types[speed] = L["%1$i%% speed %2$s"]:format( speed, L["ground"] );
                                     if( first > speed ) then
                                         first = speed;
                                     end--if
                                 end--if
                             end--for
                             for _, name in pairs( GroundPassengerMounts ) do
                                 local speed = GetSetSpeed( name );

                                 if( speed ) then
                                     local index = speed + 1000;

                                     uiTypes[index] = name;
                                     types[index] = L["%1$i%% speed %2$s"]:format( speed, L["ground passenger"] );
                                     if( first > index ) then
                                         first = index;
                                     end--if
                                 end--if
                             end--for
                             if( not types[info.handler.ui.currentType] ) then
                                 info.handler.ui.currentType = first;
                             end--if
                             return types;
                         end,
                get = function( info )
                          return info.handler.ui.currentType;
                      end,
                set = function( info, val )
                          info.handler.ui.currentType = val;
                          info.handler.ui.currentFavorite = 0;
                      end,
            },
            mount = {
                type = "select",
                order = 3,
                name = L["Mount"],
                desc = L["Mount that should always be picked for the currently selected zone and type."],
                width = "full",
                cmdHidden = true,
                values = function( info )
                             local currentType = info.handler.ui.Types[info.handler.ui.currentType] or "";

                             info.handler.ui.Favorites = {};
                             if( not info.handler.db.profile.ZoneFavorites[info.handler.ui.currentZone] ) then
                                 info.handler.db.profile.ZoneFavorites[info.handler.ui.currentZone] = {};
                             end--if

                             local Favorites = info.handler.ui.Favorites;
                             local used = {};
                             local ZoneFavorite = info.handler.db.profile.ZoneFavorites[info.handler.ui.currentZone][currentType];

                             for _, mount in pairs( info.handler:GetPetsList( currentType ) ) do
                                 table.insert( Favorites, mount:GetID() );
                             end--for
                             if( ZoneFavorite and ( not used[ZoneFavorite] ) ) then
                                 table.insert( Favorites, ZoneFavorite );
                             end--if
                             table.sort( Favorites, function( a, b ) return ( ( a < 0 ) and GetSpellInfo( a * -1 ) or GetItemInfo( a ) ) < ( ( b < 0 ) and GetSpellInfo( b * -1 ) or GetItemInfo( b ) ) end );
                             Favorites = {};
                             for index, mount in pairs( info.handler.ui.Favorites ) do
                                 local sName, sTexture;

                                 if( mount < 0 ) then
                                     sName, _, sTexture, _, _, _, _, _, _ = GetSpellInfo( mount * -1 );
                                 else
                                     sName, _, _, _, _, _, _, _, _, sTexture = GetItemInfo( mount );
                                 end--if
                                 if( sTexture ) then
                                     Favorites[index] = ("|T%s:0|t %s"):format( sTexture, sName );
                                 end--if
                                 if( ZoneFavorite == mount ) then
                                     info.handler.ui.currentFavorite = index;
                                 end--if
                             end--for
                             if( not ZoneFavorite ) then
                                 info.handler.ui.currentFavorite = 0;
                             end--if
                             Favorites[0] = L["none"];
                             return Favorites;
                         end,
                get = function( info )
                          return info.handler.ui.currentFavorite;
                      end,
                set = function( info, val )
                          info.handler.ui.currentFavorite = val;
                          info.handler.db.profile.ZoneFavorites[info.handler.ui.currentZone][uiTypes[info.handler.ui.currentType]] = info.handler.ui.Favorites[val];
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            [L["list"]] = {
                type = "execute",
                order = 4,
                guiHidden = true,
                name = L["List all zone favorite mounts"],
                desc = L["Show a list with all of your zone favorite mounts."],
                func = function( info )
                           info.handler:Print( L["List of your zone favorite mounts:"] );
                           for zone, mount in pairs( info.handler.db.profile.ZoneFavorites ) do
                               local link, sTexture;

                               if( mount > 0 ) then
                                   link = LibLordFarlander.GetItemLink( mount );
                                   _, _, _, _, _, _, _, _, _, sTexture = GetItemInfo( mount );
                               else
                                   mount = mount * -1;
                                   link = GetSpellLink( mount );
                                   _, _, sTexture, _, _, _, _, _, _ = GetSpellInfo( mount );
                               end--if
                               if( sTexture ) then
                                   info.handler:Print( ("%s: |T%s:0|t %s"):format( zone, sTexture, link or tostring( mount ) ) );
                               else
                                   info.handler:Print( ("%s: %s"):format( zone, link or tostring( mount ) ) );
                               end--if
                           end--for
                       end,
            },
        },
    };
    local modifierKeys = { ["Disabled"] = L["Disabled"], ["Alt"] = L["Alt"], ["Control"] = L["Control"], ["Shift"] = L["Shift"], };
    local options = {
        type = "group",
        handler = self,
        args = {
            [L["favorFlying"]] = {
                type = "toggle",
                order = 1,
                name = L["Favor flying mounts"],
                desc = function( info )
                            local returnString = L["Toggles if flying mounts should be favored if in an area that allows them."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.FavorFlying and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.FavorFlying;
                      end,
                set = function( info, val )
                          info.handler.db.profile.FavorFlying = val;
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            [L["favorPassenger"]] = {
                type = "toggle",
                order = 2,
                name = L["Favor passenger mounts"],
                desc = function( info )
                            local returnString = L["Toggles if mounts that can take a passenger should be favored."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.FavorPassenger and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.FavorPassenger;
                      end,
                set = function( info, val )
                          info.handler.db.profile.FavorPassenger = val;
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            [L["showModifierItems"]] = {
                type = "toggle",
                order = 3,
                name = L["Show modifier items"],
                desc = function( info )
                            local returnString = L["Toggles if items that modify mounts (such as Fresh Holly) should be selected when holding Ctrl while mounted."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.ShowModifierItems and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.ShowModifierItems;
                      end,
                set = function( info, val )
                          info.handler.db.profile.ShowModifierItems = val;
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            [L["suggestOnlySoulbound"]] = {
                type = "toggle",
                order = 4,
                name = L["Suggest only soulbound mounts"],
                desc = function( info )
                            local returnString = L["Toggles if only mounts that are soulbound to you are to be suggested for use."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.SuggestOnlySoulbound and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.SuggestOnlySoulbound;
                      end,
                set = function( info, val )
                          info.handler.db.profile.SuggestOnlySoulbound = val;
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            [L["modiferDismountFlying"]] = {
                type = "toggle",
                order = 5,
                name = L["Need modifer to dismount while flying"],
                width = "double",
                desc = function( info )
                            local returnString = L["Toggles if a modifier needs to be held down to dismount while flying."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.ModiferDismountFlying and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.ModiferDismountFlying;
                      end,
                set = function( info, val )
                          info.handler.db.profile.ModiferDismountFlying = val;
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            [L["suggestForms"]] = {
                type = "toggle",
                order = 6,
                name = L["Suggest travel forms instead of next best mount"],
                width = "double",
                desc = function( info )
                            local returnString = L["Toggles if a travel form spell will be selected instead of your next slowest mount, if availiable."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.SuggestForms and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.SuggestForms;
                      end,
                set = function( info, val )
                          info.handler.db.profile.SuggestForms = val;
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            [L["defaultToFormInCombat"]] = {
                type = "toggle",
                order = 7,
                name = L["Suggest travel form when in combat"],
                width = "double",
                desc = function( info )
                            local returnString = L["Toggles if when entering combat your travel form spell will be selected."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.DefaultToFormInCombat and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.DefaultToFormInCombat;
                      end,
                set = function( info, val )
                          info.handler.db.profile.DefaultToFormInCombat = val;
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            flyingModifier = {
                type = "select",
                order = 8,
                name = L["Modifier key for flying mounts"],
                values = modifierKeys,
                cmdHidden = true,
                desc = L["Changes the key that needs to be held down to pick a non flying mount in an area where flying is allowed."],
                width = "double",
                get = function( info )
                          return info.handler.db.profile.FlyingModifier;
                      end,
                set = function( info, val )
                          info.handler.db.profile.FlyingModifier = val;
                      end,
            },
            passengerModifier = {
                type = "select",
                order = 9,
                name = L["Modifier key for passenger mounts"],
                values = modifierKeys,
                cmdHidden = true,
                desc = L["Changes the key that needs to be held down to pick a non passenger mount."],
                width = "double",
                get = function( info )
                          return info.handler.db.profile.PassengerModifier;
                      end,
                set = function( info, val )
                          info.handler.db.profile.PassengerModifier = val;
                      end,
            },
            mountItemModifier = {
                type = "select",
                order = 10,
                name = L["Modifier key for modifier items"],
                values = modifierKeys,
                cmdHidden = true,
                desc = L["Changes the key that needs to be held down to pick a modifer item while mounted."],
                width = "double",
                get = function( info )
                          return info.handler.db.profile.MountItemModifier;
                      end,
                set = function( info, val )
                          info.handler.db.profile.MountItemModifier = val;
                      end,
            },
            nextBestModifier = {
                type = "select",
                order = 11,
                name = L["Modifier key for next slowest mount"],
                values = modifierKeys,
                cmdHidden = true,
                desc = L["Changes the key that needs to be held down to pick your next slowest mount."],
                width = "double",
                get = function( info )
                          return info.handler.db.profile.NextBestModifier;
                      end,
                set = function( info, val )
                          info.handler.db.profile.NextBestModifier = val;
                      end,
            },
            [L["list"]] = {
                type = "execute",
                order = 12,
                name = L["List all mounts"],
                desc = L["Show a list with all your mounts."],
                func = "ListMounts"
            },
            [L["refresh"]] = {
                type = "execute",
                order = 13,
                name = L["Refresh mount list"],
                desc = L["Force a refresh of the list of all your mounts."],
                func = function( info )
                           info.handler.dirtyBags = {};
                           info.handler:GetPlayersMounts();
                           info.handler:ChooseRandomMounts();
                           info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                       end,
            },
        },
        name = L["Coconuts"],
        desc = L["Addon options."],
        icon = "Interface\\AddOns\\Coconuts\\Img\\Coconuts",
    };
    
    self.db = LibStub( "AceDB-3.0" ):New( "CoconutsDB", defaults );
    self.db.RegisterCallback( self, "OnProfileChanged" );
    self.db.RegisterCallback( self, "OnProfileReset" );

    local profileOptions = LibStub( "AceDBOptions-3.0" ):GetOptionsTable( self.db );
    local slashOptions = {
        type = "group",
        handler = self,
        name = options.name,
        desc = options.desc,
        icon = options.icon,
        args = {
            [L["options"]] = options,
            [L["favorites"]] = favoritesOptions,
            [L["zoneFavorites"]] = zoneOptions,
            [L["button"]] = buttonOptions,
            [L["profile"]] = profileOptions,
            [L["gui"]] = {
                type = "execute",
                order = -1,
                name = L["Open GUI configurator"],
                desc = L["Open GUI configurator"],
                guiHidden = true,
                func = function( info )
                           InterfaceOptionsFrame_OpenToFrame( L[info.handler.name] );
                       end,
            },
            debug = {
                type = "toggle",
                hidden = true,
                name = "Debug",
                desc = "Debug mode toggle",
                desc = function( info )
                            local returnString = "Debug mode toggle";

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.DEBUG and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.DEBUG;
                      end,
                set = function( info, val )
                          info.handler.DEBUG = val;
                          info.handler:Print( ("debug mode |cFFFF0000[%s]|r"):format( info.handler.DEBUG and L["yes"] or L["no"] ) );
                      end,
            },
        },
    };

    LibStub( "AceConfig-3.0" ):RegisterOptionsTable( L["Coconuts"], slashOptions, { L["coconuts"], L["ccn"] } );
    AceConfigReg:RegisterOptionsTable( "Coconuts Main Options", options );
    AceConfigReg:RegisterOptionsTable( "Coconuts Button", buttonOptions );
    AceConfigReg:RegisterOptionsTable( "Coconuts Zone Favorites", zoneOptions );
    AceConfigReg:RegisterOptionsTable( "Coconuts Favorites", favoritesOptions );
    AceConfigReg:RegisterOptionsTable( "Coconuts Profile", profileOptions );
    AceConfigDialog:AddToBlizOptions( "Coconuts Main Options", L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( "Coconuts Button", L["Button"], L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( "Coconuts Zone Favorites", L["Zone Favorites"], L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( "Coconuts Favorites", L["Favorites"], L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( "Coconuts Profile", L["Profile"], L["Coconuts"] );
end--Coconuts:OnInitialize()

function Coconuts:OnProfileReset()
    self:OnProfileChanged();
end--Coconuts:OnProfileReset()

function Coconuts:OnProfileChanged()
    LibStub( "LibLordFarlander-UI-2.0" ).OnProfileChanged( self );
    --Upgrade check
    if( type( self.db.profile.LastVersion ) == "number" ) then
        --Upgrading from pre-wotlk, need to convert itemIDs in favorites and zone favorites to spellIDs and to the new format
        local newList = {};
        local ItemInSet = PetAndMountDatabase.ItemInSet;
        local GroundSets = PetAndMountDatabase:GetSubsets( "Mounts.Ground" );
        local FlyingSets = PetAndMountDatabase:GetSubsets( "Mounts.Flying" );        

        for zone, itemID in pairs( self.db.profile.ZoneFavorites ) do
            newList[zone] = {};
            for _, set in pairs( GroundSets ) do
                if( ItemInSet( PetAndMountDatabase, itemID, set ) ) then
                    newList[zone][set] = itemID;
                end--if
            end--if
            for _, set in pairs( FlyingSets ) do
                if( ItemInSet( PetAndMountDatabase, itemID, set ) ) then
                    newList[zone][set] = itemID;
                end--if
            end--if
        end--for
        self.db.profile.ZoneFavorites = newList;
    end--if
    if( ( not self.db.profile.LastWoWVersion ) or ( tonumber( self.db.profile.LastWoWVersion:sub( 1, 1 ) ) < tonumber( GetBuildInfo():sub( 1, 1 ) ) ) ) then
        local newList = {};
        local ItemInSet = PetAndMountDatabase.ItemInSet;

        for itemID, _ in pairs( self.db.profile.Favorites ) do
            local newID = ItemInSet( PetAndMountDatabase, itemID, "Conversion.Mounts" );

            if( newID ) then
                itemID = tonumber( newID );
            end--if
            newList[itemID] = 2.0;
        end--for
        self.db.profile.Favorites = newList;
        newList = {};
        for zone, sets in pairs( self.db.profile.ZoneFavorites ) do
            newList[zone] = {};
            if( sets ) then
                for set, itemID in pairs( sets ) do
                    local newID = ItemInSet( PetAndMountDatabase, itemID, "Conversion.Mounts" );

                    if( newID ) then
                        itemID = tonumber( newID );
                    end--if
                    newList[zone][set] = itemID;
                end--for
            end--if
        end--for
        self.db.profile.ZoneFavorites = newList;
    end--if
    self.db.profile.LastWoWVersion = GetBuildInfo();
    if( ( not self.db.profile.LastVersion ) or ( type( self.db.profile.LastVersion ) ~= "table" ) or ( self.db.profile.LastVersion.major ~= self.version.major ) or ( self.db.profile.LastVersion.minor ~= self.version.minor ) or ( self.db.profile.LastVersion.revision ~= self.version.revision ) ) then
        self.db.profile.LastVersion = self.version;
    end--if

    self.Dirty = true
    self:GetPlayersMounts();
    self:ChooseRandomMounts();
    self:SetButton( self:GetStateBasedOnModifiers() );
end--Coconuts:OnProfileChanged()

function Coconuts:OnEnable()
    --Register the Events
    local LibSpecialEvents_Mount = LibStub( "LibSpecialEvents-Mount-4.0" );
    local LibSpecialEvents_Skill = LibStub( "LibSpecialEvents-Skill-4.0" );
    local LibUI = LibStub( "LibLordFarlander-UI-2.0" );

    self:RegisterEvent( "BAG_UPDATE",
        function( _, bagID )
            if( bagID and ( bagID >= 0 ) and ( bagID <= NUM_BAG_SLOTS ) ) then
                Coconuts.dirtyBags[bagID] = true;
                Coconuts.Dirty = true;
                if( not Coconuts.button:IsVisible() ) then
                    Coconuts:ScheduleTimer( "OnUpdate", 0 );
                end--if
            end--if
        end );
    self:RegisterEvent( "MODIFIER_STATE_CHANGED",
        function()
            Coconuts:SetButton( Coconuts:GetStateBasedOnModifiers() );
        end );
    self:RegisterEvent( "PLAYER_ALIVE",
        function()
            if( not Coconuts.Loaded ) then
                Coconuts.Loaded = true;
                Coconuts.RidingSkill = LibStub( "LibSpecialEvents-Skill-4.0" ):GetSkillLevel( L["Riding"] );
            end--if
        end );
    self:RegisterEvent( "PLAYER_REGEN_DISABLED",
        function()
            if( Coconuts.db.profile.Button.AutoHideButton and ( not ( Coconuts.db.profile.DefaultToFormInCombat and Coconuts.CurrentGroundForm ) ) ) then
                Coconuts.button:HideEx();
            else
                Coconuts.button:SetFromListEntry( Coconuts.CurrentGroundForm );
            end--if
        end );
    self:RegisterEvent( "PLAYER_REGEN_ENABLED",
        function()
            if( not Coconuts.Loaded ) then
                Coconuts.Loaded = true;
                Coconuts.RidingSkill = LibStub( "LibSpecialEvents-Skill-4.0" ):GetSkillLevel( L["Riding"] );
            end--if
            if( Coconuts.db.profile.Button.AutoHideButton ) then
                Coconuts.button:ShowEx();
            end--if
            Coconuts.button:OutOfCombatReset();
            Coconuts:SetButton( Coconuts:GetStateBasedOnModifiers() );
        end );
    self:RegisterEvent( "ZONE_CHANGED_NEW_AREA",
        function()
            if( Coconuts.DEBUG and LibLordFarlander.IsInFlyableZone() ) then
                Coconuts:dprint( "Coconuts: In flyable zone" );
            end--if
            Coconuts:SetMountsForZone();
            Coconuts:SetButton( Coconuts:GetStateBasedOnModifiers() );
        end );

    LibSpecialEvents_Mount.RegisterCallback( self, "Mounted",
        function()
            if( not Coconuts.Mounted ) then
                Coconuts.Mounted = true;
                Coconuts:ChooseRandomMounts();
            end--if
        end );
    LibSpecialEvents_Mount.RegisterCallback( self, "Dismounted",
        function()
            if( Coconuts.Mounted ) then
                Coconuts.Mounted = false;
                Coconuts:SetButton( Coconuts:GetStateBasedOnModifiers() );
            end--if
        end );
    LibStub( "LibSpecialEvents-LearnSpell-4.0" ).RegisterCallback( self, "LearnedSpell",
        function( _, spell )
            local spellID = LibLordFarlander.GetItemIDFromLink( GetSpellLink( spell ) );

            --Also check for learning Cold Weather Flying
            if( ( spellID ~= nil ) and ( ( spellID == LibLordFarlander.ColdWeatherFlying ) or PetAndMountDatabase:SpellInSet( spellID, "Mounts" ) or PetAndMountDatabase:SpellInSet( spellID, "Spells.Travel" ) ) ) then
                self.Dirty = true;
                if( spellID == LibLordFarlander.ColdWeatherFlying ) then
                    Coconuts.dirtyBags = {};
                else
                    Coconuts.dirtyBags[-255] = true;
                end--if
                if( not Coconuts.button:IsVisible() ) then
                    Coconuts:ScheduleTimer( "OnUpdate", 0 );
                end--if
            end--if
        end );
    LibSpecialEvents_Skill.RegisterCallback( self, "LearnedSkill_" .. L["Riding"],
        function( _, level )
            Coconuts.RidingSkill = level;
            Coconuts:SetButton( Coconuts:GetStateBasedOnModifiers() );
        end );
    LibSpecialEvents_Skill.RegisterCallback( self, "SkillLeveled_" .. L["Riding"],
        function( _, level )
            Coconuts.RidingSkill = level;
            Coconuts:SetButton( Coconuts:GetStateBasedOnModifiers() );
        end );

    LibUI.SetupForUI( self );
    self.button = LibUI.CreateButton( self, self.name, "Interface\\AddOns\\Coconuts\\Img\\Coconuts", self.db.profile.ButtonGroup, self.SkinChanged, self );
    ItemList.AttachToButton( self.button );
    self.button:SetMovable( not self.db.profile.Button.LockButton );

    --Check if upgrading from Mount
    if( GetBindingKey( "BINDING_NAME_CLICK Mount_Button:LeftButton" ) ) then
        SetBindingClick( GetBindingKey( "BINDING_NAME_CLICK Mount_Button:LeftButton" ), self.button:GetName(), "LeftButton" );
        SetBindingClick( nil, "Mount_Button", "LeftButton" );
    end--if

    self:OnProfileChanged();
end--Coconuts:OnEnable();

function Coconuts:OnDisable()
    LibStub( "LibSpecialEvents-Mount-4.0" ).UnregisterAll( self );
    LibStub( "LibSpecialEvents-LearnSpell-4.0" ).UnregisterAll( self );
    LibStub( "LibSpecialEvents-Skill-4.0" ).UnregisterAll( self );
end--Coconuts:OnDisable()

function Coconuts:AutoHideChange()
    self:SetButton( self:GetStateBasedOnModifiers() );
end--Coconuts:AutoHideChange()

function Coconuts:OnUpdate( elapsed )
    local SwimCheck = self.CurrentAquaticForm and self.db.profile.SuggestForms;
    local oldSwimming = self.Swimming;

    self.Swimming = IsSwimming();
    if( self.Dirty ) then
        --Update the list of mount items if needed
        self.Dirty = false;        
        if( self:HasMountListChanged() ) then
            self:ChooseRandomMounts();
        end--if
        SwimCheck = false;
        self:SetButton( self:GetStateBasedOnModifiers() );
    end--if
    if( SwimCheck and ( oldSwimming ~= IsSwimming() ) ) then
        self:SetButton( self:GetStateBasedOnModifiers() );
    end--if
end--Coconuts:OnUpdate()

function Coconuts.IsModifierDown( modname )
    return ( modname ~= "Disabled" ) and _G[("Is%sKeyDown"):format( modname )]();
end--Coconuts:IsModifierDown( modname )

function Coconuts:GetStateBasedOnModifiers()
    local profile = self.db.profile;
    local IsFlyingModifierDown = self.IsModifierDown( profile.FlyingModifier );
    local IsPassengerModifierDown = self.IsModifierDown( profile.PassengerModifier );
    local IsMountItemModifierDown = self.IsModifierDown( profile.MountItemModifier );
    local IsNextBestModifierDown = self.IsModifierDown( profile.NextBestModifier );

    if( self.Mounted ) then
        if( profile.ShowModifierItems and IsMountItemModifierDown and self.CurrentMountItem ) then
            self.Mode = "Enchant";
            return self.CurrentMountItem;
        else
            self.Mode = "Dismount";
            return nil;
        end--if
    elseif( self.Swimming and self.CurrentAquaticForm and profile.SuggestForms ) then
        self.Mode = "Mount";
        return self.CurrentAquaticForm;
    elseif( ( GetRealZoneText() == L["Temple Of Ahn'Qiraj"] ) and self.TempleOfAhnQiraj ) then
        self.Mode = "Mount";
        return self.TempleOfAhnQiraj;
    else
        --Pick a mount or travel form
        if( LibLordFarlander.IsInFlyableZone() ) then
            if( ( profile.FavorFlying and ( not IsFlyingModifierDown ) ) or ( ( not profile.FavorFlying ) and IsFlyingModifierDown ) ) then
                if( ( profile.FavorPassenger and ( not IsPassengerModifierDown ) ) or ( ( not profile.FavorPassenger ) and IsPassengerModifierDown ) ) then
                    if( IsNextBestModifierDown ) then
                        if( self.NextCurrentFlyingPassenger ) then
                            self.Mode = "Mount";
                            return ( profile.SuggestForms and self.CurrentFlyingForm ) or self.NextCurrentFlyingPassenger;
                        end--if
                    else
                        if( self.CurrentFlyingPassenger ) then
                            self.Mode = "Mount";
                            return self.CurrentFlyingPassenger;
                        end--if
                    end--if
                end--if
                if( IsNextBestModifierDown ) then
                    if( self.NextCurrentFlying ) then
                        self.Mode = "Mount";
                        return ( profile.SuggestForms and self.CurrentFlyingForm ) or self.NextCurrentFlying;
                    end--if
                else
                    if( self.CurrentFlying ) then
                        self.Mode = "Mount";
                        return self.CurrentFlying;
                    end--if
                end--if
            end--if
        end--if
        if( ( profile.FavorPassenger and ( not IsPassengerModifierDown ) ) or ( ( not profile.FavorPassenger ) and IsPassengerModifierDown ) ) then
            if( IsNextBestModifierDown ) then
                if( self.NextCurrentGroundPassenger ) then
                    self.Mode = "Mount";
                    return ( profile.SuggestForms and self.CurrentGroundForm ) or self.NextCurrentGroundPassenger;
                end--if
            else
                if( self.CurrentGroundPassenger ) then
                    self.Mode = "Mount";
                    return self.CurrentGroundPassenger;
                end--if
            end--if
        end--if
        if( IsNextBestModifierDown ) then
            if( self.NextCurrentGround ) then
                self.Mode = "Mount";
                return ( profile.SuggestForms and self.CurrentGroundForm ) or self.NextCurrentGround;
            end--if
        else
            if( self.CurrentGround ) then
                self.Mode = "Mount";
                return self.CurrentGround;
            end--if
        end--if
    end--if
    self.Mode = "";
    return nil;
end--Coconuts:GetStateBasedOnModifiers()

----------------------------
--+Mount/Inventory Functions--
function Coconuts.GetSetSpeed( pt3set )
    local iterator, handle = PetAndMountDatabase:IterateSet( pt3set );
    local item, speed = iterator( handle );

    if( item and speed ) then
        speed = PetAndMountDatabase:GetMountAttributes( speed );
        return speed;
    end--if
    return nil
end--Coconuts.GetSetSpeed( pt3set )

function Coconuts:GetZoneMount( ZoneList, set )
    if( ZoneList and ( set:len() > 0 ) ) then
        ZoneList = ZoneList[set];
        if( ZoneList ) then
            if( ( ZoneList < 0 ) and LibLordFarlander.HasPlayerSpell( ZoneList * -1 ) ) then
                return( ItemList.BuildSpell( ZoneList * -1 ) );
            elseif( ( LibLordFarlander.PlayerItemCount( ZoneList ) > 0 ) and PetAndMountDatabase:ItemInSet( ZoneList, set ) ) then
                local BagItem = ItemList.FindInBags( ZoneList );
                
                if( ( not self.db.profile.SuggestOnlySoulbound ) or LibLordFarlander.ItemIsSoulbound( BagItem.bag, BagItem.slot ) ) then
                    return BagItem;
                end--if
            end--if
        end--if
    end--if
    return nil;
end--Coconuts:GetZoneMount( ZoneList, PT3set )

function Coconuts:GetRandomFromListWithWeighing( set, current, favorites )
    if( ( LibLordFarlander.count ( favorites ) > 0 ) and set and ( LibLordFarlander.count ( set ) > 1 ) ) then
        local PlayerItemCount = LibLordFarlander.PlayerItemCount;
        local GetItemIDFromLink = LibLordFarlander.GetItemIDFromLink;
        local ItemIsSoulbound = LibLordFarlander.ItemIsSoulbound;
        local HasPlayerSpell = LibLordFarlander.HasPlayerSpell;
        local ItemInSet = PetAndMountDatabase.ItemInSet;
        local opSuggestOnlySoulbound = self.db.profile.SuggestOnlySoulbound;
        local total = 0.0;
        local useset = {};
        local randomValue;
        local amountOfItems = 0;
        local index = 0;
        local item = nil;

        for _, item in pairs( set ) do
            local itemID = item:GetID();
            local howOften = favorites[itemID] or 1.0;
            local BagItem = ( itemID > 0 ) and ItemList.FindInBags( itemID ) or nil;

            if( ( howOften > 0.0 ) and ( ( ( itemID < 0 ) and HasPlayerSpell( itemID * -1 ) ) or ( ( itemID > 0 ) and ( PlayerItemCount( itemID ) > 0 ) and ( ( not opSuggestOnlySoulbound ) or ItemIsSoulbound( BagItem.bag, BagItem.slot ) ) ) ) ) then
                tinsert( useset, item );
                total = total + howOften;
                amountOfItems = amountOfItems + 1;
            end--if
        end--for
        randomValue = math.random() * total;
        repeat
            index = index + 1;
            item = useset[index];
            if( not item ) then
                self:Print( ( "useset[%i] = nil"):format( index ) );
            end
            randomValue = randomValue - ( favorites[item:GetID()] or 1.0 );
        until ( randomValue <= 0 ) or ( index == amountOfItems );
        return item;
    else
        return ItemList.GetRandomItemFromList( set, current );
    end--if
end--Coconuts:GetRandomFromListWithWeighing( masterset, set, current, favorites )

function Coconuts:SetMountsForZone()
    local ZoneList = self.db.profile.ZoneFavorites[GetRealZoneText()];

    if( ZoneList ) then
        self.CurrentGround = self:GetZoneMount( ZoneList, self.Ground.Best ) or self.CurrentGround;
        self.NextCurrentGround = self:GetZoneMount( ZoneList, self.Ground.NextBest ) or self.NextCurrentGround;
        self.CurrentFlying = self:GetZoneMount( ZoneList, self.Flying.Best ) or self.CurrentFlying;
        self.NextCurrentFlying = self:GetZoneMount( ZoneList, self.Flying.NextBest ) or self.NextCurrentFlying;
        self.CurrentGroundPassenger = self:GetZoneMount( ZoneList, self.Ground.Passenger.Best ) or self.CurrentGroundPassenger;
        self.NextCurrentGroundPassenger = self:GetZoneMount( ZoneList, self.Ground.Passenger.NextBest ) or self.NextCurrentGroundPassenger;
        self.CurrentFlyingPassenger = self:GetZoneMount( ZoneList, self.Flying.Passenger.Best ) or self.CurrentFlyingPassenger;
        self.NextCurrentFlyingPassenger = self:GetZoneMount( ZoneList, self.Flying.Passenger.NextBest ) or self.NextCurrentFlyingPassenger;
    end--if
end--Coconuts:SetMountsForZone()

function Coconuts:ChooseRandomMounts()
    local Favorites = self.db.profile.Favorites;

    self.CurrentGround = self:GetRandomFromListWithWeighing( self:GetPetsList( self.Ground.Best ), self.CurrentGround, Favorites );
    self.NextCurrentGround = self:GetRandomFromListWithWeighing( self:GetPetsList( self.Ground.NextBest ), self.NextCurrentGround, Favorites );
    self.CurrentFlying = self:GetRandomFromListWithWeighing( self:GetPetsList( self.Flying.Best ), self.CurrentFlying, Favorites );
    self.NextCurrentFlying = self:GetRandomFromListWithWeighing( self:GetPetsList( self.Flying.NextBest ), self.NextCurrentFlying, Favorites );
    
    self.CurrentGroundPassenger = self:GetRandomFromListWithWeighing( self:GetPetsList( self.Ground.Passenger.Best ), self.CurrentGroundPassenger, Favorites );
    self.NextCurrentGroundPassenger = self:GetRandomFromListWithWeighing( self:GetPetsList( self.Ground.Passenger.NextBest ), self.NextCurrentGroundPassenger, Favorites );
    self.CurrentFlyingPassenger = self:GetRandomFromListWithWeighing( self:GetPetsList( self.Flying.Passenger.Best ), self.CurrentFlyingPassenger, Favorites );
    self.NextCurrentFlyingPassenger = self:GetRandomFromListWithWeighing( self:GetPetsList( self.Flying.Passenger.NextBest ), self.NextCurrentFlyingPassenger, Favorites );

    self:SetMountsForZone();
    self:SetButton( self:GetStateBasedOnModifiers() );
end--Coconuts:ChooseRandomMounts()

function Coconuts.GetBestSets( sets )
    local Best = nil;
    local BestName = "";
    local NextBest = nil;
    local NextBestName = "";

    for name, set in pairs( sets ) do
        if( Best == nil ) then
            Best = set;
            BestName = name;
        else
            if( set.speed > Best.speed ) then
                if( ( NextBest == nil ) or ( Best.speed > NextBest.speed ) ) then
                    NextBest = Best;
                    NextBestName = BestName;
                end--if
                Best = set;
                BestName = name;
            elseif( ( NextBest == nil ) or ( set.speed > NextBest.speed ) ) then
                NextBest = set;
                NextBestName = name;
            end--if
        end--if
    end--for
    if( not NextBest ) then
        NextBestName = BestName;
    end--if
    return BestName, NextBestName;
end--Coconuts.GetBestSets()

function Coconuts.AddIfMount( ptsets, dest, bag, slot, itemID )
    for _, set in pairs( ptsets ) do
        local speed = PetAndMountDatabase:ItemInSet( itemID, set );

        if( speed ) then
            if( dest[set] == nil ) then
                dest[set] = { bags = {}, ["speed"] = PetAndMountDatabase:GetMountAttributes( speed ) };
            end--if
            if( not dest[set].bags[bag] ) then
                dest[set].bags[bag] = {};
            end--if
            ItemList.AddBagItemToList( dest[set].bags[bag], bag, slot );
            return true;
        end--if
    end--for
    return false;
end--Coconuts.AddIfMount( ptsets, dest, bag, slot, itemID )

function Coconuts.GetPlayerSpellMounts( spellDatabase, mountTable )
    local iterator, handle = PetAndMountDatabase:IterateSet( spellDatabase );

    if( iterator and handle ) then
        local spellID, speed = iterator( handle );

        while spellID do
            if( spellID < 0 ) then
                spellID = spellID * -1;
                if( LibLordFarlander.HasPlayerSpell( spellID ) ) then
                    if( mountTable.speed == nil ) then
                        mountTable.speed = PetAndMountDatabase:GetMountAttributes( speed );
                    end--if
                    if( not mountTable.bags[-255] ) then
                        mountTable.bags[-255] = {};
                    end--if
                    ItemList.AddSpellToList( mountTable.bags[-255], spellID );
                end--if
            end--if
            spellID, speed = iterator( handle );
        end--for
    end--if
end--Coconuts.GetPlayerSpellMounts( spellDatabase, mountTable )

function Coconuts.AddSpells( ptsets, dest )
    for _, set in pairs( ptsets ) do
        local destSet = dest[set];

        if( destSet == nil ) then
            destSet = { bags = {}, };
        end--if
        Coconuts.GetPlayerSpellMounts( set, destSet );
        if( destSet.speed ) then
            dest[set] = destSet;
        end--if
    end--for
end--Coconuts.AddSpells( ptsets, dest )

function Coconuts.GetTravelForm( set )
    local iterator, handle = PetAndMountDatabase:IterateSet( ("Spells.Travel.%s"):format( set ) );
    local spell, _ = iterator( handle );

    while spell do
        spell = spell * -1;
        if( LibLordFarlander.HasPlayerSpell( spell ) ) then
            return ItemList.BuildSpell( spell );
        end--if
        spell, _ = iterator( handle );
    end--for
    return nil;
end--Coconuts.GetTravelForm( set )

function Coconuts.CleanupBags( sets, bag )
    for _, set in pairs( sets ) do
        if( set.bags) then
            set.bags[bag] = {};
        end--if
    end--for
end--Coconuts.CleanupBags( set )

--Gets all your pets from your inventory.
function Coconuts:GetPlayersMounts()
    -- First do ground mountItems, best first
    local opSuggestOnlySoulbound = self.db.profile.SuggestOnlySoulbound;
    local dirtyBags = self.dirtyBags;

    if( LibLordFarlander.count( dirtyBags ) == 0 ) then
        for bag = 0, NUM_BAG_SLOTS do
            dirtyBags[bag] = true;
        end--if
        dirtyBags[-255] = true;
        dirtyBags[-254] = true;
    end--if

    self.CurrentMountItem = nil;
    self.Ground.Best = "";
    self.Ground.NextBest = "";
    self.Ground.Passenger.Best = "";
    self.Ground.Passenger.NextBest = "";
    self.Flying.Best = "";
    self.Flying.NextBest = "";
    self.Flying.Passenger.Best = "";
    self.Flying.Passenger.NextBest = "";

    local GroundMounts = PetAndMountDatabase:GetSubsets( "Mounts.Ground" ) or {};
    local FlyingMounts = PetAndMountDatabase:GetSubsets( "Mounts.Flying" ) or {};
    local GroundPassengerMounts = PetAndMountDatabase:GetSubsets( "Mounts.Passenger.Ground" ) or {};
    local FlyingPassengerMounts = PetAndMountDatabase:GetSubsets( "Mounts.Passenger.Flying" ) or {};
    local GroundSets = self.Ground.Sets;
    local FlyingSets = self.Flying.Sets;
    local GroundPassengerSets = self.Ground.Passenger.Sets;
    local FlyingPassengerSets = self.Flying.Passenger.Sets;
    local ItemIsSoulbound = LibLordFarlander.ItemIsSoulbound;
    local PlayerMeetsMountRequirements = self.PlayerMeetsMountRequirements;
    local AddIfMount = self.AddIfMount;
    local GetItemIDFromLink = LibLordFarlander.GetItemIDFromLink;

    for bag = 0, NUM_BAG_SLOTS do
        if( dirtyBags[bag] ) then
            self.CleanupBags( GroundSets, bag );
            self.CleanupBags( FlyingSets, bag );
            self.CleanupBags( GroundPassengerSets, bag );
            self.CleanupBags( FlyingPassengerSets, bag );
            for slot = 1, GetContainerNumSlots( bag ) do
                local itemLink = GetContainerItemLink( bag, slot );
    
                if( itemLink ) then
                    local itemID = GetItemIDFromLink( itemLink );

                    if( itemID and PlayerMeetsMountRequirements( self, itemID ) and ( ( not opSuggestOnlySoulbound ) or ItemIsSoulbound( bag, slot ) ) ) then
                        AddIfMount( GroundMounts, GroundSets, bag, slot, itemID );
                        AddIfMount( FlyingMounts, FlyingSets, bag, slot, itemID );
                        AddIfMount( GroundPassengerMounts, GroundPassengerSets, bag, slot, itemID );
                        AddIfMount( FlyingPassengerMounts, FlyingPassengerSets, bag, slot, itemID );
                        if( ( GetZoneText() == L["Temple Of Ahn'Qiraj"] ) and PetAndMountDatabase:ItemInSet( itemID, "TempleOfAhnQiraj.Mounts" ) ) then
                            self.TempleOfAhnQiraj = ItemList.BuildBagItem( bag, slot );
                        end--if
                    end--if
                end--if
            end--for
        end--if
    end--for

    if( dirtyBags[-255] ) then
        self.CleanupBags( GroundSets, -255 );
        self.CleanupBags( FlyingSets, -255 );
        self.CleanupBags( GroundPassengerSets, -255 );
        self.CleanupBags( FlyingPassengerSets, -255 );
        self.AddSpells( GroundMounts, GroundSets );
        self.AddSpells( FlyingMounts, FlyingSets );
        self.AddSpells( GroundPassengerMounts, GroundPassengerSets );
        self.AddSpells( FlyingPassengerMounts, FlyingPassengerSets );
    end--if

    self.Ground.Best, self.Ground.NextBest = self.GetBestSets( GroundSets );
    self.Flying.Best, self.Flying.NextBest = self.GetBestSets( FlyingSets );
    self.Ground.Passenger.Best, self.Ground.Passenger.NextBest = self.GetBestSets( GroundPassengerSets );
    self.Flying.Passenger.Best, self.Flying.Passenger.NextBest = self.GetBestSets( FlyingPassengerSets );
    
    local iterator, handle = PetAndMountDatabase:IterateSet( "Items.MountEnchant" );
    local item, _ = iterator( handle );

    while item do
        if( GetItemCount( item ) > 0 ) then
            self.CurrentMountItem = ItemList.BuildItem( item );
            break;
        end--if
        item, _ = iterator( handle );
    end--for
    
    self.CurrentGroundForm = self.GetTravelForm( "Ground" );
    self.CurrentFlyingForm = self.GetTravelForm( "Flying" );
    self.CurrentAquaticForm = self.GetTravelForm( "Aquatic" );
    self.dirtyBags = {};
end--Coconuts:GetPlayersMounts()

function Coconuts:PrintMountList( mountTable, string )
    if( #mountTable > 0 ) then
        self:Print( L["%s mounts:"]:format( string ) );
        for _, mount in pairs( mountTable ) do
            self:Print( ("|T%s:0|t %s"):format( mount:GetTexture(), mount:GetLink() ) );
        end--for
    end--if
end--Coconuts:PrintMountList()

function Coconuts:ListMountType( type, lists )
    local sorted = {};

    for name, _ in pairs( lists ) do
        tinsert( sorted, name );
    end--for
    self.CompareList = lists;
    table.sort( sorted, function( a, b ) return Coconuts.CompareList[a].speed < Coconuts.CompareList[b].speed end );
    self.CompareList = nil;
    for _, name in pairs( sorted ) do
        self:PrintMountList( self:GetPetsList( name ), L["%1$i%% speed %2$s"]:format( lists[name].speed, type ) );
    end--for
end--Coconuts.ListMountType( type, list )

--Prints out a list with all your mountItems.
function Coconuts:ListMounts()
    self:Print( L["List of your mounts:"] );

    self:ListMountType( L["ground"], self.Ground.Sets );
    self:ListMountType( L["flying"], self.Flying.Sets );
    self:ListMountType( L["ground passenger"], self.Ground.Passenger.Sets );
    self:ListMountType( L["flying passenger"], self.Flying.Passenger.Sets );
end--Coconuts:ListMounts()

function Coconuts:GetPetsList( set )
    local returnVal = {};

    if( set:find( ".Passenger" ) ) then
        if( set:find( ".Ground" ) ) then
            if( self.Ground.Passenger.Sets[set] and self.Ground.Passenger.Sets[set].bags ) then
                for _, bag in pairs( self.Ground.Passenger.Sets[set].bags ) do
                    for _, mount in pairs( bag ) do
                        tinsert( returnVal, mount );
                    end--for
                end--for
            end--if
            return returnVal;
        end--if
        if( set:find( ".Flying" ) ) then
            if( self.Flying.Passenger.Sets[set] and self.Flying.Passenger.Sets[set].bags ) then
                for _, bag in pairs( self.Flying.Passenger.Sets[set].bags ) do
                    for _, mount in pairs( bag ) do
                        tinsert( returnVal, mount );
                    end--for
                end--for
            end--if
            return returnVal;
        end--if
    end--if
    if( set:find( ".Ground" ) ) then
        if( self.Ground.Sets[set] and self.Ground.Sets[set].bags ) then
            for _, bag in pairs( self.Ground.Sets[set].bags ) do
                for _, mount in pairs( bag ) do
                    tinsert( returnVal, mount );
                end--for
            end--for
        end--if
        return returnVal;
    end--if
    if( set:find( ".Flying" ) ) then
        if( self.Flying.Sets[set] and self.Flying.Sets[set].bags ) then
            for _, bag in pairs( self.Flying.Sets[set].bags ) do
                for _, mount in pairs( bag ) do
                    tinsert( returnVal, mount );
                end--for
            end--for
        end--if
        return returnVal;
    end--if
    return returnVal;
end--Coconuts.GetPetsList( set )

function Coconuts.GetSetMountsList( sets, returnList, used )
    if( not used ) then
        used = {};
    end--if
    for _, set in pairs( sets ) do
        for _, bag in pairs( set.bags ) do
            for _, mount in pairs( bag ) do
                if( not used[mount:GetID()] ) then
                    tinsert( returnList, mount );
                    used[mount:GetID()] = true;
                end--if
            end--for
        end--for
    end--for
end--Coconuts.GetSetMountsList( set, returnList, used )

function Coconuts:GetAllMountsList()
    local returnList = {};
    local used = {};
    
    self.GetSetMountsList( self.Ground.Sets, returnList, used );
    self.GetSetMountsList( self.Flying.Sets, returnList, used );
    self.GetSetMountsList( self.Ground.Passenger.Sets, returnList, used );
    self.GetSetMountsList( self.Flying.Passenger.Sets, returnList, used );
    return returnList;
end--Coconuts:GetAllMountsList()
---Coconuts/Inventory Functions--
----------------------------

-------------------
--+Help Functions--
function Coconuts:PlayerMeetsMountRequirements( itemID )
    local speed = PetAndMountDatabase:ItemInSet( itemID, "Mounts" );

    if( not self.RidingSkill ) then
        self.RidingSkill = LibStub( "LibSpecialEvents-Skill-4.0" ):GetSkillLevel( L["Riding"] );
    end--if
    if( speed ) then
        local _, _, _, _, itemMinLevel, _, _, _, _, _ = GetItemInfo( itemID );    
        local _, ridingSkill = PetAndMountDatabase:GetMountAttributes( speed );

        if( itemMinLevel == nil ) then
            itemMinLevel = 0;
        end--if
        return ( itemMinLevel <= UnitLevel( "player" ) ) and ( ridingSkill <= self.RidingSkill );
    end--if
    return false;
end--if

function Coconuts:HasMountListChanged()
    local selfBestGround = self.Ground.Best and self.Ground.Sets[self.Ground.Best] or {};
    local selfBestFlying = self.Flying.Best and self.Flying.Sets[self.Flying.Best] or {};
    local selfBestPassengerGround = self.Ground.Passenger.Best and self.Ground.Passenger.Sets[self.Ground.Passenger.Best] or {};
    local selfBestPassengerFlying = self.Flying.Passenger.Best and self.Flying.Passenger.Sets[self.Flying.Passenger.Best] or {};
    local selfNextBestGround = self.Ground.NextBest and self.Ground.Sets[self.Ground.NextBest] or {};
    local selfNextBestFlying = self.Flying.NextBest and self.Flying.Sets[self.Flying.NextBest] or {};
    local selfNextBestPassengerGround = self.Ground.Passenger.NextBest and self.Ground.Passenger.Sets[self.Ground.Passenger.NextBest] or {};
    local selfNextBestPassengerFlying = self.Flying.Passenger.NextBest and self.Flying.Passenger.Sets[self.Flying.Passenger.NextBest] or {};
    local preGroundCount = #selfBestGround;
    local preFlyingCount = #selfBestFlying;
    local prePassengerGroundCount = #selfBestPassengerGround;
    local prePassengerFlyingCount = #selfBestPassengerFlying;
    local preNextGroundCount = #selfNextBestGround;
    local preNextFlyingCount = #selfNextBestFlying;
    local preNextPassengerGroundCount = #selfNextBestPassengerGround;
    local preNextPassengerFlyingCount = #selfNextBestPassengerFlying;
    local preBestGround =  {};
    local preBestFlying = {};
    local preBestPassengerGround =  {};
    local preBestPassengerFlying = {};
    local preNextBestGround =  {};
    local preNextBestFlying = {};
    local preNextBestPassengerGround =  {};
    local preNextBestPassengerFlying = {};

    LibLordFarlander.CopyTable( preBestGround, selfBestGround );
    LibLordFarlander.CopyTable( preBestFlying, selfBestFlying );
    LibLordFarlander.CopyTable( preBestPassengerGround, selfBestPassengerGround );
    LibLordFarlander.CopyTable( preBestPassengerFlying, selfBestPassengerFlying );
    LibLordFarlander.CopyTable( preNextBestGround, selfNextBestGround );
    LibLordFarlander.CopyTable( preNextBestFlying, selfNextBestFlying );
    LibLordFarlander.CopyTable( preNextBestPassengerGround, selfNextBestPassengerGround );
    LibLordFarlander.CopyTable( preNextBestPassengerFlying, selfNextBestPassengerFlying );

    self:GetPlayersMounts();
    
    selfBestGround = self.Ground.Best and self.Ground.Sets[self.Ground.Best] or {};
    selfBestFlying = self.Flying.Best and self.Flying.Sets[self.Flying.Best] or {};
    selfBestPassengerGround = self.Ground.Passenger.Best and self.Ground.Passenger.Sets[self.Ground.Passenger.Best] or {};
    selfBestPassengerFlying = self.Flying.Passenger.Best and self.Flying.Passenger.Sets[self.Flying.Passenger.Best] or {};
    selfNextBestGround = self.Ground.NextBest and self.Ground.Sets[self.Ground.NextBest] or {};
    selfNextBestFlying = self.Flying.NextBest and self.Flying.Sets[self.Flying.NextBest] or {};
    selfNextBestPassengerGround = self.Ground.Passenger.NextBest and self.Ground.Passenger.Sets[self.Ground.Passenger.NextBest] or {};
    selfNextBestPassengerFlying = self.Flying.Passenger.NextBest and self.Flying.Passenger.Sets[self.Flying.Passenger.NextBest] or {};

    if( not self.ReadyToInitialize ) then
        self.ReadyToInitialize = true;
        return true;
    end
    self.ReadyToInitialize = true;
    return not( LibLordFarlander.CompareTables( preBestGround, selfBestGround ) and LibLordFarlander.CompareTables( preBestFlying, selfBestFlying ) and
        LibLordFarlander.CompareTables( preNextBestGround, selfNextBestGround ) and LibLordFarlander.CompareTables( preNextBestFlying, selfNextBestFlying ) and
        LibLordFarlander.CompareTables( preBestPassengerGround, selfBestPassengerGround ) and LibLordFarlander.CompareTables( preBestPassengerFlying, selfBestPassengerFlying ) and
        LibLordFarlander.CompareTables( preNextBestPassengerGround, selfNextBestPassengerGround ) and LibLordFarlander.CompareTables( preNextBestPassengerFlying, selfNextBestPassengerFlying ) );
end--Coconuts:HasMountListChanged()
---Help Functions--
-------------------

-----------------
--+UI Functions--
function Coconuts:SetGameTooltips()
    local mntButton = self.button;
    local mntTooltip = mntButton.tooltip;

    if( ( not self.keyBoundMode ) and ( not InCombatLockdown() ) ) then
        local mountItem = self.CurrentMount;
        local chardb = self.db.profile;

        mntTooltip:SetOwner( mntButton, "RIGHT" );
        if( self.Mode == "Dismount" ) then
            if( chardb.ModiferDismountFlying ) then
                if( chardb.ShowModifierItems ) then
                    local notUsed = {};
                    local used = self.db.profile.MountItemModifier;
                    
                    if( used ~= "Alt" ) then
                        tinsert( notUsed, L["Alt"] );
                    end--if
                    if( used ~= "Control" ) then
                        tinsert( notUsed, L["Control"] );
                    end--if
                    if( used ~= "Shift" ) then
                        tinsert( notUsed, L["Shift"] );
                    end--if
                    mntTooltip:SetText( L["Dismount (if flying hold down %s or %s and click to dismount)"]:format( notUsed[1], notUsed[2] ) );
                else
                    mntTooltip:SetText( L["Dismount (if flying hold down Shift, Control, or Alt and click to dismount)"] );
                end--if
            else
                mntTooltip:SetText( L["Dismount"] );
            end--if
        elseif( ( self.Mode == "Mount" ) or ( self.Mode == "Enchant" ) ) then
            mntTooltip:SetFromListEntry( mountItem );
        else
            mntTooltip:SetText( L["You have no mounts."] );
        end--if
    elseif( InCombatLockdown() ) then
        mntTooltip:SetText( L["Cannot mount, you are in combat!"] );
    end--if
    LibStub( "LibLordFarlander-UI-2.0" ).AfterSetTooltip( mntButton );
end--Coconuts:SetGameTooltips()

function Coconuts:SetButton( mountItem )
    local mntButton = self.button;
    local chardb = self.db.profile;

    if( not InCombatLockdown() ) then
        if( chardb.Button.AutoHideButton and mountItem ) then
            mntButton:ShowEx();
        end--if
        if( self.Mode == "Dismount" ) then
            mntButton.cooldown:Hide();
            mntButton:ShowEx();
            mntButton:Enable();
            mntButton:SetAttribute( "*type1", "macro" );
            mntButton:SetAttribute( "macrotext", chardb.ModiferDismountFlying and "/dismount [noflying][mod]" or "/dismount" );
            mntButton:SetImage( "Interface\\AddOns\\Coconuts\\Img\\Dismount" );
        elseif( mountItem and ( ( self.Mode == "Mount" ) or ( self.Mode == "Enchant" ) ) ) then
            mntButton:Enable();
            mntButton:SetFromListEntry( mountItem, not LibLordFarlander.CompareTables( self.CurrentMount, mountItem ) );
        else
            if( chardb.Button.AutoHideButton ) then
                mntButton:HideEx();
            end--if
            mntButton.cooldown:Hide();
            mntButton:SetAttribute( "*type1", "item" );
            mntButton:SetAttribute( "*bag1", nil );
            mntButton:SetAttribute( "*slot1", nil );
            mntButton:SetAttribute( "*item1", nil );
            mntButton:SetImage( "Interface\\AddOns\\Coconuts\\Img\\Coconuts" );
        end--if
        self.CurrentMount = mountItem;
    end--if
end--Coconuts:SetButton( mountItem )
---UI Functions--
-----------------
